
/******************************************************************************
* MODULE     : tft.hpp
* DESCRIPTION: generic low-level TFT multiplication
* COPYRIGHT  : (C) 2005  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX__TFT__HPP
#define __MMX__TFT__HPP
#include "tft_tree.hpp"

namespace mmx {
#define TMPL template<class C>
#define TMPL_DEF template<typename C>
#define Tft_tree tft_tree<C>
#define Tft_tree_rep tft_tree_rep<C>
#define Tft_leaf tft_leaf<C>
#define Tft_binary tft_binary<C>

/******************************************************************************
* Univariate direct TFT
******************************************************************************/

TMPL inline void
fft_partial (const fft_transformer<C>& ffter, C* dest, nat d, nat offset) {
  nat size= 1 << d;
  nat n   = 1 << ffter.steps;
  nat step= ffter.steps - d;
  for (; step<ffter.steps; step++) {
    nat w= n >> (1+step);
    nat l= ffter.order[step];
    if (w*l < size) fft_step (dest, size, w, ffter.a, l);
    else fft_step (dest, size, w, ffter.a + ((offset/w) % l), size/w);
  }
}

TMPL inline void
fft_cross_range (C* lp, C* rp, const fft_root<C>* a, nat plus, nat nr) {
  if (plus == 0)
    for (nat i=nr; i!=0; i--) {
      fft_cross (lp, rp);
      lp++; rp++;
    }
  else
    for (nat i=nr; i!=0; i--) {
      fft_cross (lp, rp, a+plus);
      lp++; rp++;
    }
}

TMPL void
tft (const fft_transformer<C>& ffter, C* dest, nat s, nat d, nat offset) {
  while (s != 0) {
    nat w= 1 << (d-1);
    nat step= ffter.steps - d;
    nat l   = ffter.order[step];
    nat plus= (offset/w) % l;
    fft_cross_range (dest, dest+w, ffter.a, plus, w);
    if (s >= w) {
      fft_partial (ffter, dest, d-1, offset);
      offset += w;
      dest += w;
      s -= w;
    }
    d--;
  }
}

TMPL void
tft (C* a, nat s, nat d) {
  if (s == ((nat) 1) << d) fft (a, s, false);
  else {
    fft_transformer<C> ffter (1 << d);
    segment_clear (a + s, (1 << d) - s);
    tft (ffter, a, s, d, 0);
  }
}

// TODO: different source and destination sizes

/******************************************************************************
* Univariate inverse TFT
******************************************************************************/

TMPL inline void
ifft_partial (const fft_transformer<C>& ffter, C* dest, nat d, nat offset) {
  nat size = 1 << d;
  nat n    = 1 << ffter.steps;
  nat first= ffter.steps - d;
  nat last = ffter.steps;
  for (nat step= last-1; step!=first-1; step--) {
    nat w= n >> (1+step);
    nat l= ffter.order[step];
    if (w*l < size) ifft_step (dest, size, w, ffter.a + 1, l);
    else ifft_step (dest, size, w, ffter.a + 1 + ((offset/w) % l), size/w);
  }
  fft_shift (dest, size);
}

TMPL inline void
itft_flip_range (C* lp, C* rp, const fft_root<C>* a, nat plus, nat nr) {
  if (plus == 0)
    for (nat i=nr; i!=0; i--) {
      itft_flip (lp, rp);
      lp++; rp++;
    }
  else
    for (nat i=nr; i!=0; i--) {
      itft_flip (lp, rp, a+plus);
      lp++; rp++;
    }
}

TMPL inline void
ifft_cross_range (C* lp, C* rp, const fft_root<C>* a, nat plus, nat nr) {
  static C h= invert (C ((int) 2));
  if (plus == 0)
    for (nat i=nr; i!=0; i--) {
      ifft_cross (lp, rp);
      *lp= (*lp) * h;
      *rp= (*rp) * h;
      lp++; rp++;
    }
  else
    for (nat i=nr; i!=0; i--) {
      ifft_cross (lp, rp, a+plus+1);
      *lp= (*lp) * h;
      *rp= (*rp) * h;
      lp++; rp++;
    }
}

TMPL void
itft (const fft_transformer<C>& ffter, C* dest, nat s, nat d, nat offset) {
  //cout << "Itft " << offset << " -- " << offset+s << " (" << d << ")\n";
  nat w= 1 << (d-1);
  if (s < w) {
    nat step= ffter.steps - d;
    nat l   = ffter.order[step];
    nat plus= (offset/w) % l;
    fft_cross_range (dest+s, dest+s+w, ffter.a, plus, w-s);
    itft (ffter, dest, s, d-1, offset);
    itft_flip_range (dest, dest+w, ffter.a, plus, s);
  }
  else {
    ifft_partial (ffter, dest, d-1, offset);
    nat step= ffter.steps - d;
    nat l   = ffter.order[step];
    nat plus= (offset/w) % l;
    itft_flip_range (dest+s-w, dest+s, ffter.a, plus, (w<<1)-s);
    if (s == w) return;
    itft (ffter, dest+w, s-w, d-1, offset+w);
    ifft_cross_range (dest, dest+w, ffter.a, plus, s-w);
  }
}

TMPL void
itft (C* a, nat s, nat d) {
  if (s == ((nat) 1) << d) fft (a, s, true);
  else {
    fft_transformer<C> ffter (1 << d);
    segment_clear (a + s, (1 << d) - s);
    itft (ffter, a, s, d, 0);
  }
}

/******************************************************************************
* The direct TFT
******************************************************************************/

TMPL void
tft_cross (Tft_tree& lt, nat li, Tft_tree& rt, nat ri,
	   fft_root<C>* a, nat plus) {
  //cout << "    Cross " << *a << "\n";
  assert (li == 0 || ri == 0);
  if (is_zero (lt) && is_zero (rt)) return;
  else if (is_zero (lt)) {
    if (is_atomic (rt)) expand_zero (lt);
    else split (lt);
  }
  else if (is_zero (rt)) {
    if (is_atomic (lt)) expand_zero (rt);
    else split (rt);
  }

  if (is_atomic (lt) && is_atomic (rt))
    fft_cross_range (seg (lt) + li, seg (rt) + ri,
		     a, plus, min (N(lt), N(rt)));
  else if (is_atomic (lt)) {
    nat h= N(rt)>>1;
    tft_cross (lt, li  , car (rt), 0, a, plus);
    tft_cross (lt, li+h, cdr (rt), 0, a, plus);
  }
  else if (is_atomic (rt)) {
    nat h= N(lt)>>1;
    tft_cross (car (lt), 0, rt, ri  , a, plus);
    tft_cross (cdr (lt), 0, rt, ri+h, a, plus);
  }
  else {
    tft_cross (car (lt), 0, car (rt), 0, a, plus);
    tft_cross (cdr (lt), 0, cdr (rt), 0, a, plus);
  }
}

TMPL void
tft (const fft_transformer<C>& ffter, Tft_tree& t, nat offset) {
  //cout << "  Tft " << offset << "--" << offset+N(t) << "\n";
  //print (t, 4); cout << "\n";
  if (!is_target (t)) {
    t= Tft_leaf (D(t)); // Save space
    //cout << "    Killed\n";
    return;
  }
  if (is_atomic (t)) {
    if (is_zero (t)) return;
    fft_partial (ffter, seg (t), D(t), offset);
  }
  else {
    nat step= ffter.steps - D(t);
    nat l   = ffter.order[step];
    nat w   = N(t) >> 1;
    tft_cross (car (t), 0, cdr (t), 0, ffter.a, (offset/w) % l);
    // permutation here allows for earlier removal of unused space
    tft (ffter, cdr (t), offset + (N(t)>>1));
    tft (ffter, car (t), offset);
  }
  //cout << "  Result " << offset << "--" << offset+N(t) << "\n";
  //print (t, 4); cout << "\n";
}

/*
TMPL fft_transformer<C>
bivariate (const Tft_tree& t) {
  nat i=1;
  for (; i<N(t); i*=2)
    if (i*i == N(t)) break;
  assert (i*i == N(t));
  nat order[2];
  order[0]= order[1]= i;
  return fft_transformer<C> (order, 2);
}
*/

TMPL void
tft (Tft_tree& t) {
  fft_transformer<C> ffter (N(t));
  //fft_transformer<C> ffter= bivariate (t);
  tft (ffter, t, 0);
}

/******************************************************************************
* The inverse TFT
******************************************************************************/

TMPL void
itft_down (Tft_tree& lt, nat li, Tft_tree& rt, nat ri,
	   fft_root<C>* a, nat plus)
{
  //cout << "    Down " << *a << "\n";
  assert (li == 0 || ri == 0);
  if (!is_forward (lt) || !is_forward (rt)) return;
  if (is_zero (lt) && is_zero (rt)) return;
  else if (is_zero (lt)) {
    if (is_atomic (rt)) expand_zero (lt);
    else split (lt);
  }
  else if (is_zero (rt)) {
    if (is_atomic (lt)) expand_zero (rt);
    else split (rt);
  }

  if (is_atomic (lt) && is_atomic (rt))
    fft_cross_range (seg (lt) + li, seg (rt) + ri,
		     a, plus, min (N(lt), N(rt)));
  else if (is_atomic (lt)) {
    nat h= N(rt)>>1;
    itft_down (lt, li  , car (rt), 0, a, plus);
    itft_down (lt, li+h, cdr (rt), 0, a, plus);
  }
  else if (is_atomic (rt)) {
    nat h= N(lt)>>1;
    itft_down (car (lt), 0, rt, ri  , a, plus);
    itft_down (cdr (lt), 0, rt, ri+h, a, plus);
  }
  else {
    itft_down (car (lt), 0, car (rt), 0, a, plus);
    itft_down (cdr (lt), 0, cdr (rt), 0, a, plus);
  }
}

TMPL void
itft_flip (Tft_tree& lt, nat li, Tft_tree& rt, nat ri,
	   fft_root<C>* a, nat plus)
{
  //cout << "    Flip " << *a << "\n";
  assert (li == 0 || ri == 0);
  if (!is_backward (lt) || !is_forward (rt)) return;
  if (is_zero (lt) && is_zero (rt)) return;
  else if (is_zero (lt)) {
    if (is_atomic (rt)) expand_zero (lt);
    else split (lt);
  }
  else if (is_zero (rt)) {
    if (is_atomic (lt)) expand_zero (rt);
    else split (rt);
  }

  if (is_atomic (lt) && is_atomic (rt))
    itft_flip_range (seg (lt) + li, seg (rt) + ri,
		     a, plus, min (N(lt), N(rt)));
  else if (is_atomic (lt)) {
    nat h= N(rt)>>1;
    itft_flip (lt, li  , car (rt), 0, a, plus);
    itft_flip (lt, li+h, cdr (rt), 0, a, plus);
  }
  else if (is_atomic (rt)) {
    nat h= N(lt)>>1;
    itft_flip (car (lt), 0, rt, ri  , a, plus);
    itft_flip (cdr (lt), 0, rt, ri+h, a, plus);
  }
  else {
    itft_flip (car (lt), 0, car (rt), 0, a, plus);
    itft_flip (cdr (lt), 0, cdr (rt), 0, a, plus);
  }
}

TMPL void
itft_up (Tft_tree& lt, nat li, Tft_tree& rt, nat ri,
	 fft_root<C>* a, nat plus)
{
  //cout << "    Up " << *a << "\n";
  assert (li == 0 || ri == 0);
  if (!is_backward (lt) || !is_backward (rt)) return;
  /* FIXME: check whether this code is really superfluous
  if (is_zero (lt) && is_zero (rt)) return;
  else if (is_zero (lt)) {
    if (is_atomic (rt)) expand_zero (lt);
    else split (lt);
  }
  else if (is_zero (rt)) {
    if (is_atomic (lt)) expand_zero (rt);
    else split (rt);
  }
  */

  if (is_atomic (lt) && is_atomic (rt))
    ifft_cross_range (seg (lt) + li, seg (rt) + ri,
		      a, plus, min (N(lt), N(rt)));
  else if (is_atomic (lt)) {
    nat h= N(rt)>>1;
    itft_up (lt, li  , car (rt), 0, a, plus);
    itft_up (lt, li+h, cdr (rt), 0, a, plus);
  }
  else if (is_atomic (rt)) {
    nat h= N(lt)>>1;
    itft_up (car (lt), 0, rt, ri  , a, plus);
    itft_up (cdr (lt), 0, rt, ri+h, a, plus);
  }
  else {
    itft_up (car (lt), 0, car (rt), 0, a, plus);
    itft_up (cdr (lt), 0, cdr (rt), 0, a, plus);
  }
}

TMPL void
itft (const fft_transformer<C>& ffter, Tft_tree& t, nat offset) {
  //cout << "  Itft " << offset << "--" << offset+N(t) << "\n";
  //print (t, 4); cout << "\n";
  if (!is_backward (t)) {
    t= Tft_leaf (D(t)); // Save space
    //cout << "    Killed\n";
    return;
  }
  if (is_atomic (t)) {
    if (is_zero (t)) return;
    ifft_partial (ffter, seg (t), D(t), offset);
  }
  else {
    nat last = ffter.steps;
    nat first= last - D(t);
    nat l    = ffter.order[first];
    nat w    = N(t) >> 1;
    nat plus = (offset/w) % l;
    itft_down (car (t), 0, cdr (t), 0, ffter.a, plus);
    itft (ffter, car (t), offset);
    itft_flip (car (t), 0, cdr (t), 0, ffter.a, plus);
    itft (ffter, cdr (t), offset + (N(t)>>1));
    itft_up (car (t), 0, cdr (t), 0, ffter.a, plus);
  }
  //cout << "  Result " << offset << "--" << offset+N(t) << "\n";
  //print (t, 4); cout << "\n";
}

TMPL void
itft (Tft_tree& t) {
  fft_transformer<C> ffter (N(t));
  //fft_transformer<C> ffter= bivariate (t);
  itft (ffter, t, 0);
}

/******************************************************************************
* The direct pyramidal TFT
******************************************************************************/

TMPL Tft_tree&
operator += (Tft_tree& t, Tft_tree& u) {
  assert (D(t) == D(u));
  if (is_zero (u)) return t;
  assert (!is_zero (t));
  if (is_atomic (t) && !is_atomic (u)) split (t);
  if (is_atomic (u) && !is_atomic (t)) split (u);
  if (is_atomic (t) && is_atomic (u)) {
    segment_add (seg (t), seg (u), N(t));
#ifdef FFT_COUNT
    cross_count += N(t);
#endif
    return t;
  }
  else {
    car (t) += car (u);
    cdr (t) += cdr (u);
    return t;
  }
}

TMPL Tft_tree
copy_forward (const Tft_tree& t) {
  if (is_atomic (t)) {
    Tft_tree r= is_zero (t)? Tft_leaf (D(t)): Tft_leaf (D(t), seg (t));
    r->status.is_target = false;
    r->status.is_forward= true;
#ifdef FFT_COUNT
    if (!is_zero (t)) cross_count += N(t);
#endif
    return r;
  }
  else return Tft_binary (copy_forward (car (t)), copy_forward (cdr (t)));
}

TMPL void
tft_twist (Tft_tree& t, Tft_tree& u) {
  assert (D(t) == D(u));
  if (is_zero (t)) t= copy_forward (u);
  else {
    assert (!is_zero (u));
    if (is_atomic (t) && !is_atomic (u)) split (t);
    if (is_atomic (u) && !is_atomic (t)) split (u);
    if (is_atomic (t) && is_atomic (u)) {
      segment_mul (seg (t), seg (t), C(-2), N(t));
      segment_add (seg (t), seg (u), N(t));
#ifdef FFT_COUNT
      cross_count += N(t);
#endif
    }
    else {
      tft_twist (car (t), car (u));
      tft_twist (cdr (t), cdr (u));
      t->status.is_target = is_target  (car (t)) | is_target  (cdr (t));
      t->status.is_forward= is_forward (car (t)) | is_forward (cdr (t));
    }
  }
}

TMPL Tft_tree&
car (Tft_tree& t, nat nr) {
  if (nr == 0) return t;
  else return car (car (t), nr-1);
}

TMPL void
pyramidal_tft_loop (const fft_transformer<C>& ffter,
		    Tft_tree& sum, Tft_tree& t, nat dim) {
  if (dim == 0) tft (ffter, t, 0);
  else {
    Tft_tree& u= car (cdr (t), dim-1);
    tft_twist (u, sum);
    tft (ffter, u, N(t) >> 1);
    pyramidal_tft_loop (ffter, sum, car (t), dim-1);
  }
}

TMPL void
pyramidal_tft (Tft_tree& t, nat dim) {
  fft_transformer<C> ffter (N(t));
  //fft_transformer<C> ffter= bivariate (t);
  Tft_tree& sum= car (t, dim);
  for (nat i=0; i<dim; i++)
    sum += car (cdr (car (t, i)), dim-i-1);
  pyramidal_tft_loop (ffter, sum, t, dim);
}

/******************************************************************************
* The inverse pyramidal TFT
******************************************************************************/

TMPL Tft_tree&
operator -= (Tft_tree& t, Tft_tree& u) {
  assert (D(t) == D(u));
  if (is_zero (u)) return t;
  assert (!is_zero (t));
  if (is_atomic (t) && !is_atomic (u)) split (t);
  if (is_atomic (u) && !is_atomic (t)) split (u);
  if (is_atomic (t) && is_atomic (u)) {
    segment_sub (seg (t), seg (u), N(t));
#ifdef FFT_COUNT
    cross_count += N(t);
#endif
    return t;
  }
  else {
    car (t) -= car (u);
    cdr (t) -= cdr (u);
    return t;
  }
}

TMPL void
pyramidal_itft_down (Tft_tree& t, Tft_tree& u) {
  assert (D(t) == D(u));
  if (is_zero (u));
  else if (is_zero (t)) t= copy_forward (u);
  else {
    if (is_atomic (t) && !is_atomic (u)) split (t);
    if (is_atomic (u) && !is_atomic (t)) split (u);
    if (!is_atomic (t) && !is_atomic (u)) {
      pyramidal_itft_down (car (t), car (u));
      pyramidal_itft_down (cdr (t), cdr (u));
    }
  }
}

TMPL void
itft_twist (Tft_tree& t, Tft_tree& u) {
  assert (D(t) == D(u));
  if (!is_backward (t)) t= Tft_leaf (D(t)); // Save space
  else {
    assert (!is_zero (u));
    if (is_atomic (t) && !is_atomic (u)) split (t);
    if (is_atomic (u) && !is_atomic (t)) split (u);
    if (is_atomic (t) && is_atomic (u)) {
      segment_sub (seg (t), seg (u), N(t));
      segment_div (seg (t), seg (t), C(-2), N(t));
#ifdef FFT_COUNT
      flip_count += N(t);
#endif
    }
    else {
      itft_twist (car (t), car (u));
      itft_twist (cdr (t), cdr (u));
      t->status.is_target = is_target  (car (t)) | is_target  (cdr (t));
      t->status.is_forward= is_forward (car (t)) | is_forward (cdr (t));
    }
  }
}

TMPL void
pyramidal_itft_loop (const fft_transformer<C>& ffter,
		    Tft_tree& sum, Tft_tree& t, nat dim) {
  if (dim == 0) itft (ffter, t, 0);
  else {
    pyramidal_itft_loop (ffter, sum, car (t), dim-1);
    Tft_tree& u= car (cdr (t), dim-1);
    pyramidal_itft_down (u, sum);
    itft (ffter, u, N(t) >> 1);
    itft_twist (u, sum);
  }
}

TMPL void
pyramidal_itft (Tft_tree& t, nat dim) {
  fft_transformer<C> ffter (N(t));
  //fft_transformer<C> ffter= bivariate (t);
  Tft_tree& sum= car (t, dim);
  pyramidal_itft_loop (ffter, sum, t, dim);
  for (nat i=0; i<dim; i++)
    sum -= car (cdr (car (t, i)), dim-i-1);
}

#undef TMPL_DEF
#undef TMPL
#undef Tft_tree
#undef Tft_tree_rep
#undef Tft_leaf
#undef Tft_binary
} // namespace mmx
#endif //__MMX__TFT__HPP
