
/******************************************************************************
* MODULE     : file.cpp
* DESCRIPTION: Loading and saving files
* COPYRIGHT  : (C) 2000  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include <basix_utils.hpp>
#include <fstream>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
using std::ifstream;
using std::ofstream;

namespace mmx {

string sys_home_path;
string sys_mmx_path;
string sys_load_path;

/******************************************************************************
* System commands
******************************************************************************/

string
get_env (const string& var) {
  char* _var= as_charp (var);
  char* _ret= getenv (_var);
  free_charp (_var);
  if (_ret==NULL) return "";
  string ret (_ret);
  return ret;
  // do not delete _ret !
}

int
system (const string& s) {
  char* _s= as_charp (s);
  int r= std::system (_s);
  free_charp (_s);
  return r;
}

void
init_system () {
  if (sys_home_path == "") {
    sys_home_path= get_env ("HOME");
    string sys_mmx_path= get_env ("MMX_PATH");
    if (sys_mmx_path == "") sys_mmx_path= sys_home_path * "/.mathemagix";
    system ("mkdir -p " * sys_mmx_path);
    system ("mkdir -m 744 -p " * sys_mmx_path * "/lib");
    system ("mkdir -m 744 -p " * sys_mmx_path * "/mmx");
    system ("mkdir -m 744 -p " * sys_mmx_path * "/system");
    sys_load_path= get_env ("MMX_LOAD_PATH");
    if (sys_load_path == "") sys_load_path= sys_mmx_path * "/mmx";
  }
}

string
eval_system (const string& s) {
  init_system ();
  string temp= sys_mmx_path * "/system/temp";
  system (s * " > " * temp);
  string r;
  bool flag= load (temp, r, false);
  system ("rm -rf " * temp);
  if (flag) return "";
  while ((N(r)>0) && (r[N(r)-1]=='\n')) r= r (0, N(r)-1);
  return r;
}

/******************************************************************************
* Routines for file names
******************************************************************************/

bool
file_exists (const string& file_name) {
  char* _file_name= as_charp (file_name);
  ifstream f (_file_name);
  free_charp (_file_name);
  if (!f) return false;
  return true;
}

string
file_in_path (const string& file_path, const string& file_name) {
  nat i, n= N(file_path), start= 0;
  for (i=0; i<=n; i++)
    if ((i==n) || (file_path[i] == ':')) {
      string ss= file_path (start, i);
      if ((N(ss) > 0) && (ss[0] == '$')) {
	init_system ();
	if (ss == "$MMX_LOAD_PATH") ss= sys_load_path;
	else ss= get_env (ss (1, N(ss)));
	string r= file_in_path (ss, file_name);
	if (r != "") return r;
      }
      else {
	if (ss == "") ss= ".";
	if (ss[N(ss)-1] != '/') ss= ss * "/";
	if (file_exists (ss * file_name)) return ss * file_name;
      }
      start= i+1;
    }
  return "";
}

string
get_dir_name (const string& name) {
  for (int i= N(name)-1; i>=0; i--)
    if (name[i]=='/')
      return name (0, i);
  return '.';
}

string
get_file_name (const string& name) {
  for (int i= N(name)-1; i>=0; i--)
    if (name[i]=='/')
      return name (i+1, N(name));
  return name;
}

/******************************************************************************
* Loading and saving
******************************************************************************/

bool
load (const string& file_name, string& s, bool decode) {
  if (decode) return load (eval_system ("echo " * file_name), s, false);
  char* _file_name= as_charp (file_name);
  ifstream f (_file_name);
  free_charp (_file_name);
  if (!f) return true;
  else {
    char c;
    s= string ("");
    while (f.get (c)) s << c;
    // insert_source (s, file_name);
    return false;
  }
}

bool
load (const string& file_path, const string& file_name,
      string& s, bool decode)
{
  if (decode)
    return load (file_path, eval_system ("echo " * file_name), s, false);
  if ((N(file_name) > 0) && (file_name[0] == '/'))
    return load (file_name, s, false);
  nat i, n= N(file_path), start= 0;
  for (i=0; i<=n; i++)
    if ((i==n) || (file_path[i] == ':')) {
      string ss= file_path (start, i);
      if ((N(ss) > 0) && (ss[0] == '$')) {
	init_system ();
	if (ss == "$MMX_LOAD_PATH") ss= sys_load_path;
	else ss= get_env (ss (1, N(ss)));
	if (!load (ss, file_name, s, false)) return false;
      }
      else {
	if (ss == "") ss= ".";
	if (ss[N(ss)-1] != '/') ss= ss * "/";
	if (!load (ss * file_name, s, false)) return false;
      }
      start= i+1;
    }
  return true;
}

bool
save (const string& file_name, const string& s, bool decode) {
  if (decode) return save (eval_system ("echo " * file_name), s, false);
  char* _file_name= as_charp (file_name);
  ofstream f (_file_name);
  free_charp (_file_name);
  if (!f) return true;
  else {
    nat i, n=N(s);
    for (i=0; f && (i<n); i++) f.put (s[i]);
    return false;
  }
}

} // namespace mmx
